(*^

::[paletteColors = 128; 
	fontset = title, "Times-Bold", 24, L2, center, bold, nohscroll;
	fontset = subtitle, "Times-Bold", 18, L2, center, bold, nohscroll;
	fontset = subsubtitle, "Times-Bold", 14, L2, center, bold, nohscroll;
	fontset = section, "Times-Bold", 18, L2, bold, nohscroll, grayBox;
	fontset = subsection, "Times-Bold", 14, L2, bold, nohscroll, blackBox;
	fontset = subsubsection, "Times-Bold", 12, L2, bold, nohscroll, whiteBox;
	fontset = text, "Times-Roman", 15, L2, nohscroll;
	fontset = smalltext, "Times-Roman", 13, L2, nohscroll;
	fontset = input, "Courier-Bold", 12, L2, bold, nowordwrap;
	fontset = output, "Courier", 12, L2, nowordwrap;
	fontset = message, "Courier", 13, L2, R21845, G21845, B21845, nowordwrap;
	fontset = print, "Courier", 13, L2, nowordwrap;
	fontset = info, "Courier", 13, L2, nowordwrap;
	fontset = postscript, "Courier", 24, L2, nowordwrap;
	fontset = name, "Times-Italic", 12, L2, italic, R21845, G21845, B21845, nowordwrap, nohscroll;
	fontset = header, "Times", 10, L2;
	fontset = footer, "Times", 12, L2, center;
	fontset = help, "Times-Roman", 14, L2, nohscroll;
	fontset = clipboard, "New York", 12, L2;
	fontset = completions, "Courier", 16, L2, nowordwrap;
	fontset = network, "Courier", 10, L2, nowordwrap;
	fontset = graphlabel, "Courier", 12, L2, nowordwrap;
	fontset = special1, "New York", 12, L2, nowordwrap;
	fontset = special2, "New York", 12, L2, center, nowordwrap;
	fontset = special3, "New York", 12, L2, right, nowordwrap;
	fontset = special4, "New York", 12, L2, nowordwrap;
	fontset = special5, "New York", 12, L2, nowordwrap;]
:[font = input; initialization; ]
*)

BeginPackage["ImplicitRungeKutta`"]
(* *********************************************************************
  @(#) ImplicitRungeKutta.ma,  Release 1 (11/1/90). Copyright: Terry Feagin (UHCL)

ImplicitRungeKutta integrates system of First Order ODEs by an embedded
Explicit Runge-Kutta method of order (4)5 due to Dormand and Prince. It is
best for tolerances 10^-4 to 10^-7.  The step-size is dynamically adjusted
based upon the local error tolerance `errorTolerance'. Use this package for a
Stiff system of ODEs. [ Given a system of ODEs y' = f, if the ratio of
the largest and smallest eigenvalues of the matrix (partial f/partial y)  is
very large the system is Stiff. Stiff ODEs arise in chemical kinetics. Stiffness
can cause severe limiting of the time step used during integration of the system.]

For example, given the system of differential equations:

	y' == y - yz
	z' == -z + yz

to integrate the system from x=0 to x=1, error tolerance=1.0e-5, initial 
values  y(0)=z(0)=.5 and initial step size of 0.1:

	ImplicitRungeKutta[ {y - y z, -z + y z}, {x,y,z}, {0,.5,.5}, 1, 10^-5, InitialStepSize -> 0.1]

The formal syntax is:

	ImplicitRungeKutta[ListofODEexpressions, ListOfVariables, ListOfInitialValues, 
		Finalx, ErrorTolerance, (options...)]

Where ListofODEexpressions is a list of ODE expressions; ListOfVariables is a list
that begins with the independent variable followed by the dependent variables in
the same order as the ODE expressions were specified in the ListOfODEexpressions;
ListOfInitialValues contains the initial values for ALL the variables specified
in the ListOfVariables, i.e., including the independent variable; Finalx is value
of the independent variable the solution is desired at and ErrorTolerance is the
local error tolerance to be met at each step of the intergration.

ImplicitRungeKutta has a total of four optional parameters. IRKWorkingPrecision, IRKInitialStepSize,
IRKMaximumStepSize and IRKProgressTrace. The first three are self explanatory; the
fourth causes ImplicitRungeKutta to print the current x and step size (h), and local error as it
integrates the system of ODEs.  It also prints a message when a step is
rejected because the error tolerance was not satisfied. For example,

	ImplicitRungeKutta[ {y - y z, -z + y z}, {x,y,z}, {0,.5,.5}, 1, 10^-5,
			IRKWorkingPrecision -> 20, IRKInitialStepSize -> 0.1,
			IRKMaximumStepSize -> 0.2, IRKProgressTrace -> True ]

The default values for these optinal parameters are:

	IRKWorkingPrecision -> Precision[N[1]], 
	IRKMaximumStepSize -> Finalx-Initialx, 
    IRKInitialStepSize -> IRKMaximumStepSize,
    IRKProgressTrace -> False.

If the solution is desired at a number of intermediate points, set the
IRKMaximumStepSize equal to the interval between two of these intermediate
points.

ImplicitRungeKutta package also provides the function `PlotODESolution' to plot the results 
for any one component in the system of ODEs.  PlotODESolution returns a graphics
object which can be redisplayed via `Show'. For example,

	In[2] =	ImplicitRungeKutta[ {y - y z, -z + y z}, {x,y,z}, {0,.5,.5}, 6, 10^-5]
	Out[2] = ....
	In[3] = PlotODESolution[ %,2,3,PlotRange{ {0,8},{0,5}} ]   (* plot y vs. z *)


****************************************************************** *)

ImplicitRungeKutta::usage = "ImplicitRungeKutta[ListofODEexpressions, ListOfVariables, ListOfInitialValues, 
Finalx, ErrorTolerance, (OptionalParameters...)] integrates system of ODEs
by an embedded explicit Runge-Kutta method of order 4-5. E.g., to solve
	y' == y - yz
	z' == -z + yz
		ImplicitRungeKutta[ {y - y z, -z + y z}, {x,y,z}, {0,.5,.5}, 1, 10^-5,\n
		IRKInitialStepSize->0.1, IRKProgressTrace->True ]
	ListOfVariables contains the independent variable followed by dependent
	variables in the same order as the ODEs. Initial values are specified in the same
	order. The solution can be plotted by invoking the function `PlotODESolution' for any
	given component, i.e., x vs y, x vs z. Do ?PlotODESolution for details."

Options[ImplicitRungeKutta] := {	IRKWorkingPrecision -> Precision[N[1]], 
			IRKInitialStepSize -> Automatic,
			IRKMaximumStepSize -> Automatic, 
			IRKProgressTrace -> False
		     }

Begin["`private`"]

ImplicitRungeKutta[f_List, vars_List, initialvalues_List, finalx_, errorTolerance_, opts___] :=
         RK45[f, vars, initialvalues, finalx, errorTolerance,
                        IRKWorkingPrecision /. {opts} /. Options[ImplicitRungeKutta],
                        IRKInitialStepSize /. {opts} /. Options[ImplicitRungeKutta],
                        IRKMaximumStepSize /. {opts} /. Options[ImplicitRungeKutta],
                        IRKProgressTrace /. {opts} /. Options[ImplicitRungeKutta] ]


RK45[f_, vars_, initialvalues_, finalx_, errorTolerance_,
     IRKWorkingPrecision_Integer,IRKInitialStepSize_,
     IRKMaximumStepSize_, IRKProgressTrace_ ] := 
    Block[
         {  
	    x = N[First[initialvalues], IRKWorkingPrecision], xname = vars[[1]],
	    y = N[Rest[initialvalues], IRKWorkingPrecision], 
	    xend = N[finalx, IRKWorkingPrecision], h, err, hnew, y4, y5, solution, hmax,
	    k = Table[ i+j, {i,3}, {j, Length[vars]-1}], (* dummy values *)
            c = N[ { 0, 1/2, 1}, IRKWorkingPrecision],
            a = N[ { {0, 0}, {1/4, 1/4}, {0, 1}}, IRKWorkingPrecision],
	    b = N[ {0, 1, 0}, IRKWorkingPrecision],
	    d = N[{1/6, 2/3, 1/6}, IRKWorkingPrecision],
	    tenth = N[1/10, IRKWorkingPrecision], five = N[5,IRKWorkingPrecision], 
	    ninetenth = N[9/10, IRKWorkingPrecision],

	    (* 	Make a function out of the ODE expression list. This saves time
	 	over f/. Thread[ vars->vals ] type of evaluation.
	    *)
	    ode = Function[ Release[vars], Release[f]]
	 }, 
	
	(* Check to make sure that formal paramters are of the right type *)
	If[ !NumberQ[xend], 
	    Print[StringForm["ImplicitRungeKutta::wrongargtype Final `` is not a number", xname]]; Return[{}]];
	If[ !NumberQ[errorTolerance], 
	    Print["ImplicitRungeKutta::wrongargtype ErrorTolerance is not a number"]; Return[{}]];

	If[ SameQ[IRKMaximumStepSize,Automatic], 
	  (* then *) 
		hmax = xend-x, 
	  (* else *) 
		hmax = N[IRKMaximumStepSize, IRKWorkingPrecision];
		If[ !NumberQ[hmax], 
	    		Print["ImplicitRungeKutta::wrongargtype IRKMaximumStepSize is not a number"]; 
			Return[{}]]
	];
	If[ SameQ[IRKInitialStepSize,Automatic], 
	  (* then *) 
		h = hmax, 
	  (* else *) 
		h = N[IRKInitialStepSize, IRKWorkingPrecision];
		If[ !NumberQ[h], 
	    		Print["ImplicitRungeKutta::wrongargtype IRKInitialStepSize is not a number"]; 
			Return[{}]]
	];

	(* we have to keep x separate because the k's are added to dependent
           varaibles only. That is why we all these Flattens.
        *)
        solution = {Flatten[{x,y}]};
        derk = Table[ 0, {j,Length[vars]-1}];
	While[ x < xend,
	  If[x+h == x, 
	    Print[StringForm["ImplicitRungeKutta::progress Step size too small: (``+stepsize == ``), `` = ``, h = ``, Abort further steps.", xname, xname, xname, x, h]]; Break[]];
	  If[x+h-xend > 0, h=xend-x, ];

          (* Apply the pure function `ode' we created above. Apply replaces
	     the head of {x,y} which is `List' by the head `ode'. Thus the
	     expression list `f' provided by the user becomes the function
             body of the pure function ode. This is faster than saying:

	         k[[1]] = f /. Thread[ vars-> Flatten[{x, y}]];
          *)

	  k[[1]] = Apply[ ode, Flatten[{x, y}]];
	  k[[2]] = k[[1]] + (h/2) derk;   
	  oldk = k[[2]];
	  (*
	  Print["Testing oldk and k[[1]]"];
	  Print[oldk,k[[1]]];
	  *)
	  change = 10 errorTolerance;
	  iter = 0;
	  While[(change > errorTolerance)&&(iter < 5),
	  	k[[2]] = Apply[ ode, Flatten[{x+ c[[2]] h, y+ h a[[2]].Take[k,2]}]];
	  	change = h L2norm[k[[1]], k[[2]] ,oldk];	
	 	oldk = k[[2]];
	 	iter = iter + 1; 
	 	];
	  k[[3]] = Apply[ ode, Flatten[{x+ c[[3]] h, y+ h a[[3]].Take[k,2]}]];	  
	  derk = (k[[3]] - k[[2]])/(h/2);
	  y4 = y + h b.k;
	  y5 = y + h d.k;
			(* Estimate local error *)
	  err = L2norm[y, y4,y5];

          If[IRKProgressTrace,
	    Print[StringForm["ImplicitRungeKutta::progress  x = ``, step = ``, local error = ``", x, h, err]]];

	  (* Computation of next step size. We'll require .1 <= hnew/h <=5 *)
	  hnew = Min[ hmax, h/Max[ tenth, Min[five,((err/errorTolerance)^(1/3))/ninetenth] ]];
	  If[err <= 2 errorTolerance,
	    (*then step is accepted*) 
	      x += h; y = y5; AppendTo[solution, Flatten[{x,y5}]],
	    (*else step is rejected *)
	      hnew=Min[hnew,h];
	      If[IRKProgressTrace, 
		Print[StringForm["ImplicitRungeKutta::progress Step rejected, new step-size ``", hnew ]]];

	  ];
	  h = hnew
	];
	solution
    ]

(* 
 Scaled L2 norm
*)
L2norm[y_, y4_, y5_] := 
   Sqrt[Sum[((y5[[i]]-y4[[i]])/Max[Abs[y[[i]]],Abs[y4[[i]]],1])^2, {i,Length[y]}]/Length[y]]

End[]  (* private *)

EndPackage[ ]
(*
:[font = input; ]
InputForm[Flatten[N[ImplicitRungeKutta[{-10^3 s e+10^2 x,-10^3 s e+10^2 x+10 x,10^3 s e-10^2 x-10 x,10 x},
	{t,s,e,x,p},{0,10,1,0 ,0},.01,10^-3,IRKMaximumStepSize->.01],6]]]
^*)