
BeginPackage["RungeKutta`"]

(* *********************************************************************
  @(#) RungeKutta.m,  Release 1 (2/10/89). Copyright WRI 1988. (Author: Sohail Aslam).

RungeKutta integrates system of First Order ODEs by an embedded
Explicit Runge-Kutta method of order (4)5 due to Dormand and Prince. It is
best for tolerances 10^-4 to 10^-7.  The step-size is dynamically adjusted
based upon the local error tolerance `errorTolerance'. Use this package for
Nonstiff system of ODEs. [ Given a system of ODEs y' = f, if the ratio of
the largest and smallest eigenvalues of the matrix (partial f/partial y)  is
very large the system is Stiff. Stiff ODEs arise in chemical kinetics. Stiffness
can cause severe limiting of the time step used during integration of the system.]

For example, given the system of differential equations:

	y' == y - yz
	z' == -z + yz

to integrate the system from x=0 to x=1, error tolerance=1.0e-5, initial 
values  y(0)=z(0)=.5 and initial step size of 0.1:

	RungeKutta[ {y - y z, -z + y z}, {x,y,z}, {0,.5,.5}, 1, 10^-5, InitStepSize -> 0.1]

The formal syntax is:

	RungeKutta[ListofODEexpressions, ListOfVariables, ListOfInitialValues, 
		Finalx, ErrorTolerance, (options...)]

Where ListofODEexpressions is a list of ODE expressions; ListOfVariables is a list
that begins with the independent variable followed by the dependent variables in
the same order as the ODE expressions were specified in the ListOfODEexpressions;
ListOfInitialValues contains the initial values for ALL the variables specified
in the ListOfVariables, i.e., including the independent variable; Finalx is value
of the independent variable the solution is desired at and ErrorTolerance is the
local error tolerance to be met at each step of the intergration.

RungeKutta has a total of four optinal parameters. WorkingPrecision, InitialStepSize,
MaximumStepSize and ProgressTrace. The first three are self explanatory; the
fourth causes RungeKutta to print the current x and step size (h), and local error as it
integrates the system of ODEs.  It also prints a message when a step is
rejected because the error tolerance was not satisfied. For example,

	RungeKutta[ {y - y z, -z + y z}, {x,y,z}, {0,.5,.5}, 1, 10^-5,
			WorkingPrecision -> 20, InitialStepSize -> 0.1,
			MaximumStepSize -> 0.2, ProgressTrace -> True ]

The default values for these optinal parameters are:

	WorkingPrecision -> Precision[N[1]], 
	MaximumStepSize -> Finalx-Initialx, 
        InitialStepSize -> MaximumStepSize,
        ProgressTrace -> False.

If the solution is desired at a number of intermediate points, set the
MaximumStepSize equal to the interval between two of these intermediate
points.

RungeKutta package also provides the function `PlotODESolution' to plot the results 
for any one component in the system of ODEs.  PlotODESolution returns a graphics
object which can be redisplayed via `Show'. For example,

	In[2] =	RungeKutta[ {y - y z, -z + y z}, {x,y,z}, {0,.5,.5}, 6, 10^-5]
	Out[2] = ....
	In[3] = PlotODESolution[ %,2,3,PlotRange{ {0,8},{0,5}} ]   (* plot y vs. z *)


****************************************************************** *)

RungeKutta::usage = "RungeKutta[ListofODEexpressions, ListOfVariables, ListOfInitialValues, 
Finalx, ErrorTolerance, (OptionalParameters...)] integrates system of ODEs
by an embedded explicit Runge-Kutta method of order 4-5. E.g., to solve\n
\n\ty' == y - yz\n
\tz' == -z + yz\n
\n\n\tRungeKutta[ {y - y z, -z + y z}, {x,y,z}, {0,.5,.5}, 1, 10^-5,\n
\t\tInitialStepSize->0.1, ProgressTrace->True ]\n
\nListOfVariables contains the independent variable followed by dependent
variables in the same order as the ODEs. Initial values are specified in the same
order. The solution can be plotted by invoking the function `PlotODESolution' for any
given component, i.e., x vs y, x vs z. Do ?PlotODESolution for details."

Options[RungeKutta] := {	WorkingPrecision -> Precision[N[1]], 
			InitialStepSize -> Automatic,
			MaximumStepSize -> Automatic, 
			ProgressTrace -> False
		     }

PlotODESolution::usage = "`PlotODESolution[ SolutionList, i, j, options___) ]' plots
Table[SolutionList[[k,i]],SolutionList[[k,j]], {k, Length[SolutionList]}]. 
Options are passed on to ListPlot."


Begin["`private`"]

RungeKutta[f_List, vars_List, initialvalues_List, finalx_, errorTolerance_, opts___] :=
         RK45[f, vars, initialvalues, finalx, errorTolerance,
                        WorkingPrecision /. {opts} /. Options[RungeKutta],
                        InitialStepSize /. {opts} /. Options[RungeKutta],
                        MaximumStepSize /. {opts} /. Options[RungeKutta],
                        ProgressTrace /. {opts} /. Options[RungeKutta] ]


RK45[f_, vars_, initialvalues_, finalx_, errorTolerance_, WorkingPrecision_Integer, 
	InitialStepSize_, MaximumStepSize_, ProgressTrace_ ] := 
    Block[
         {  
	    x = N[First[initialvalues], WorkingPrecision], xname = vars[[1]],
	    y = N[Rest[initialvalues], WorkingPrecision], 
	    xend = N[finalx, WorkingPrecision], h, err, hnew, y4, y5, solution, hmax,
	    k = Table[ i+j, {i,7}, {j, Length[vars]-1}], (* dummy values *)
            c = N[ { 0, 1/5, 3/10, 4/5, 8/9, 1,1}, WorkingPrecision],
            a = N[ { {0}, {1/5}, {3/40,9/40}, {44/45,-56/15,32/9},
		  {19372/6561,-25360/2187,64448/6561,-212/729},
		  {9017/3168,-355/33,46732/5247,49/176,-5103/18656},
		  {35/384,0,500/1113,125/192,-2187/6784,11/84}}, WorkingPrecision],
	    b = N[ {35/384,0,500/1113,125/192,-2187/6784,11/84,0}, WorkingPrecision],
	    d = N[{5179/57600,0,7571/16695,393/640,-92097/339200,187/2100,1/40}, WorkingPrecision],
	    tenth = N[1/10, WorkingPrecision], five = N[5,WorkingPrecision], 
	    ninetenth = N[9/10, WorkingPrecision],

	    (* 	Make a function out of the ODE expression list. This saves time
	 	over f/. Thread[ vars->vals ] type of evaluation.
	    *)
	    ode = Function[ Release[vars], Release[f]]
	 }, 
	
	(* Check to make sure that formal paramters are of the right type *)
	If[ !NumberQ[xend], 
	    Print[StringForm["RungeKutta::wrongargtype Final `` is not a number", xname]]; Return[{}]];
	If[ !NumberQ[errorTolerance], 
	    Print["RungeKutta::wrongargtype ErrorTolerance is not a number"]; Return[{}]];

	If[ SameQ[MaximumStepSize,Automatic], 
	  (* then *) 
		hmax = xend-x, 
	  (* else *) 
		hmax = N[MaximumStepSize, WorkingPrecision];
		If[ !NumberQ[hmax], 
	    		Print["RungeKutta::wrongargtype MaximumStepSize is not a number"]; 
			Return[{}]]
	];
	If[ SameQ[InitialStepSize,Automatic], 
	  (* then *) 
		h = hmax, 
	  (* else *) 
		h = N[InitialStepSize, WorkingPrecision];
		If[ !NumberQ[h], 
	    		Print["RungeKutta::wrongargtype InitialStepSize is not a number"]; 
			Return[{}]]
	];

	(* we have to keep x separate because the k's are added to dependent
           varaibles only. That is why we all these Flattens.
        *)
        solution = {Flatten[{x,y}]};
	While[ x < xend,
	  If[x+h == x, 
	    Print[StringForm["RungeKutta::progress Step size too small: (``+stepsize == ``), `` = ``, h = ``, Abort further steps.", xname, xname, xname, x, h]]; Break[]];
	  If[x+h-xend > 0, h=xend-x, ];

          (* Apply the pure function `ode' we created above. Apply replaces
	     the head of {x,y} which is `List' by the head `ode'. Thus the
	     expression list `f' provided by the user becomes the function
             body of the pure function ode. This is faster than saying:

	         k[[1]] = f /. Thread[ vars-> Flatten[{x, y}]];
          *)

	  k[[1]] = Apply[ ode, Flatten[{x, y}]];
	  k[[2]] = Apply[ ode, Flatten[{x+ c[[2]] h, y+ h a[[2,1]] k[[1]]}]];
	  k[[3]] = Apply[ ode, Flatten[{x+ c[[3]] h, y+ h a[[3]].Take[k,2]}]];
	  k[[4]] = Apply[ ode, Flatten[{x+ c[[4]] h, y+ h a[[4]].Take[k,3]}]];
	  k[[5]] = Apply[ ode, Flatten[{x+ c[[5]] h, y+ h a[[5]].Take[k,4]}]];
	  k[[6]] = Apply[ ode, Flatten[{x+ c[[6]] h, y+ h a[[6]].Take[k,5]}]];
	  k[[7]] = Apply[ ode, Flatten[{x+ c[[7]] h, y+ h a[[7]].Take[k,6]}]];

	  y4 = y + h b.k;
	  y5 = y + h d.k;


	  (* Estimate local error *)
	  err = L2norm[y, y4,y5];

          If[ProgressTrace,
	    Print[StringForm["RungeKutta::progress  `` = ``, step = ``, local error = ``", xname, x, h, err]]];

	  (* Computation of next step size. We'll require .1 <= hnew/h <=5 *)
	  hnew = Min[ hmax, h/Max[ tenth, Min[five,((err/errorTolerance)^(1/5))/ninetenth] ]];
	  If[err <= errorTolerance,
	    (*then step is accepted*) 
	      x += h; y = y4; AppendTo[solution, Flatten[{x,y4}]],
	    (*else step is rejected *)
	      hnew=Min[hnew,h];
	      If[ProgressTrace, 
		Print[StringForm["RungeKutta::progress Step rejected, new step-size ``", hnew ]]];

	  ];
	  h = hnew
	];
	solution
    ]

(* 
 Scaled L2 norm
*)
L2norm[y_, y4_, y5_] := 
   Sqrt[Sum[((y5[[i]]-y4[[i]])/Max[Abs[y[[i]]],Abs[y4[[i]]],1])^2, {i,Length[y]}]/Length[y]]

End[]  (* private *)

PlotODESolution[ SolutionList_List, i_Integer, j_Integer, options___ ] :=
	ListPlot[Table[ {SolutionList[[k,i]], SolutionList[[k,j]]}, {k,1,Length[SolutionList]}],
		   PlotJoined -> True, options ]

PlayODESolution[ SolutionList_List, i_Integer, j_Integer, options___ ] :=
	ListPlay[Table[ {SolutionList[[k,i]], SolutionList[[k,j]]}, {k,1,Length[SolutionList]}]]

EndPackage[ ]



