(*^

::[paletteColors = 128; 
	fontset = title, "Times-Bold", 24, L2, center, bold, nohscroll;
	fontset = subtitle, "Times-Bold", 18, L2, center, bold, nohscroll;
	fontset = subsubtitle, "Times-Bold", 14, L2, center, bold, nohscroll;
	fontset = section, "Times-Bold", 18, L2, bold, nohscroll, grayBox;
	fontset = subsection, "Times-Bold", 14, L2, bold, nohscroll, blackBox;
	fontset = subsubsection, "Times-Bold", 12, L2, bold, nohscroll, whiteBox;
	fontset = text, "Times-Roman", 15, L2, nohscroll;
	fontset = smalltext, "Times-Roman", 13, L2, nohscroll;
	fontset = input, "Courier-Bold", 12, L2, bold, nowordwrap;
	fontset = output, "Courier", 12, L2, nowordwrap;
	fontset = message, "Courier", 13, L2, R21845, G21845, B21845, nowordwrap;
	fontset = print, "Courier", 13, L2, nowordwrap;
	fontset = info, "Courier", 13, L2, nowordwrap;
	fontset = postscript, "Courier", 24, L2, nowordwrap;
	fontset = name, "Times-Italic", 12, L2, italic, R21845, G21845, B21845, nowordwrap, nohscroll;
	fontset = header, "Times", 10, L2;
	fontset = footer, "Times", 12, L2, center;
	fontset = help, "Times-Roman", 14, L2, nohscroll;
	fontset = clipboard, "New York", 12, L2;
	fontset = completions, "Courier", 16, L2, nowordwrap;
	fontset = network, "Courier", 10, L2, nowordwrap;
	fontset = graphlabel, "Courier", 12, L2, nowordwrap;
	fontset = special1, "New York", 12, L2, nowordwrap;
	fontset = special2, "New York", 12, L2, center, nowordwrap;
	fontset = special3, "New York", 12, L2, right, nowordwrap;
	fontset = special4, "New York", 12, L2, nowordwrap;
	fontset = special5, "New York", 12, L2, nowordwrap;]
:[font = input; initialization; pageBreak; ]
*)
BeginPackage["TaylorSeries`"]

(* *********************************************************************
  @(#) TaylorSeries.ma,  Release 1 (11/10/90). Copyright: Michael J. Mezzino, Jr. (UHCL)
*)

Options[TaylorSeries] := {	
			TSWorkingPrecision -> Precision[N[1]], 
			TSInitialStepSize -> Automatic,
			TSMaximumStepSize -> Automatic, 
			TSProgressTrace -> False
		    }

TaylorSeries::usage = "TaylorSeries[ListofODEexpressions, ListOfVariables, ListOfInitialValues, 
Finalx, ErrorTolerance, OrderOfSeries, (OptionalParameters...)] integrates system of ODEs
by using the Taylor Series expansion of order = OrderOfSeries. E.g., to solve\n\n
\ty' == y - yz
\tz' == -z + yz\n\n
TaylorSeries[ {y - y z, -z + y z}, {x,y,z}, {0,.5,.5}, 1, 10^-5,2,\n
TSInitialStepSize->0.1, TSProgressTrace->True ]\n
ListOfVariables contains the independent variable followed by dependent
variables in the same order as the ODEs. Initial values are specified in the same
order. The solution can be plotted by invoking the function `PlotODESolution' for any
given component, i.e., x vs y, x vs z. Do ?PlotODESolution for details."

Begin["`private`"]

TaylorSeries[f_List,vars_List,initialvalues_List,finalx_,
	         errorTolerance_,order_Integer?Positive,opts___] :=
    TSn[f, vars, initialvalues, finalx, errorTolerance, order,
       TSWorkingPrecision /. {opts} /. Options[TaylorSeries],
       TSInitialStepSize /. {opts} /. Options[TaylorSeries],
       TSMaximumStepSize /. {opts} /. Options[TaylorSeries],
       TSProgressTrace /. {opts} /. Options[TaylorSeries] ]

 TSn[f_, vars_, initialvalues_, finalx_, errorTolerance_,order_,
     TSWorkingPrecision_Integer,TSInitialStepSize_,
     TSMaximumStepSize_, TSProgressTrace_ ] := 
 Block[
     {
       i,j,l,
	   point = ToRules[LogicalExpand[vars==initialvalues]],		
	   fold = ToRules[LogicalExpand[Rest[Dt[vars,vars[[1]]]]==f]],
       xstep = N[First[initialvalues],TSWorkingPrecision],
       ystep = N[Rest[initialvalues],TSWorkingPrecision],
       xend = N[finalx,TSWorkingPrecision],
       k  = kk = Table[i+j,{i,order},{j,Length[vars]-1}],
       kerror1 = kerror2 = Table[i+j,{i,order},{j,Length[vars]-1}],
       b  = N[Table[i,{i,order}],TSWorkingPrecision],
       tenth = N[1/10,TSWorkingPrecision],
	   five = N[5,TSWorkingPrecision], 
	   ninetenth = N[9/10, TSWorkingPrecision],
	   h,hmax,ke1,ke2,be1,be2,err,
       solution
     },

	 (* Check to make sure that formal parameters are of the right type *)
	 If[ !NumberQ[xend], 
	    Print[StringForm["TaylorSeries::wrongargtype Final `` is not a number", vars[[1]]]];
	    Return[{}]
	 ];

	 If[ !NumberQ[errorTolerance], 
	    Print["TaylorSeries::wrongargtype ErrTol is not a number"];
	    Return[{}]
	 ];

	 If[ SameQ[TSMaximumStepSize,Automatic], 
	  (* then *) 
		hmax = xend-xstep, 
	  (* else *) 
		hmax = N[TSMaximumStepSize, TSWorkingPrecision];
		If[ !NumberQ[hmax], 
	    	Print["TaylorSeries::wrongargtype TSMaximumStepSize is not a number"]; 
			Return[{}]]
	 ];

	 If[ SameQ[TSInitialStepSize,Automatic], 
	  (* then *) 
		h = hmax, 
	  (* else *) 
		h = N[TSInitialStepSize, TSWorkingPrecision];
		If[ !NumberQ[h], 
	    	Print["TaylorSeries::wrongargtype TSInitialStepSize is not a number"]; 
			Return[{}]]
	 ];

     solution = {Flatten[{xstep,ystep}]};
	 
	 For[kk[[1]] = f ; b[[1]] = h ; i=2 , i<=order , i++,
	 	 kk[[i]] = Dt[kk[[i-1]],vars[[1]]]/.fold;
	 ];	 
	 kerror1 = Dt[kk[[order]],vars[[1]]]/.fold;
	 kerror2 = Dt[kerror1,vars[[1]]]/.fold;

     While[ xstep < xend,         
	     If[xstep+h == xstep, 
	         Print[StringForm["TaylorSeries::progress Step size too small:
	           (``+stepsize == ``), `` = ``, h = ``,
	           Abort further steps.", xstep, h]];
	           Break[]
	     ];

	     For[b[[1]] = h ; i=2 , i<=order , i++,
             b[[i]] = b[[i-1]]*h/i
	     ];	 
		 be1 = b[[order]]*h/(order+1);
		 be2 = be1*h/(order+2);
		 ke1 = kerror1/.point;
		 ke2 = kerror2/.point;
		 err = TSNorm[be1 ke1,be2 ke2];

	     (* Next step size. We'll require .1 <= hnew/h <=5 *)
	     h = Min[ hmax, h/Max[ tenth,
	             Min[five,((err/errorTolerance)^(1/(order+1)))/ninetenth]]];

	 	 If [TSProgressTrace,
	         Print[StringForm["TaylorSeries::progress Step size: err = `` , hnew = ``",err,h]]
	     ];

		 k = kk/.point;
         xstep += h;
         ystep += b.k;
	     point = ToRules[LogicalExpand[vars == Flatten[{xstep,ystep}]]];		
         AppendTo[solution,Flatten[{xstep,ystep}]]
     ];
     solution
  ]
  
TSNorm[ye1_,ye2_] := Max[Sqrt[Sum[ye1[[l]]^2,{l,Length[ye1]}]],
						 Sqrt[Sum[ye2[[l]]^2,{l,Length[ye2]}]] ]
End[]

EndPackage[]
(*
:[font = input; ]
(*
:[font = input; ]
TaylorSeries[{y,-x},{t,x,y},{0,1,0},1,10^-2,1,
			TSInitialStepSize->.1,
			TSMaximumStepSize->.1,
			TSProgressTrace->True]
:[font = input; ]
*)
^*)